"""Command constants and classes."""

from __future__ import annotations
from enum import Enum
from typing import TypedDict, Callable

from elro.utils import (
    get_default,
    get_device_names,
    set_device_name,
    get_device_states,
)

COUNT = "count"
ACK_APP = "APP_answer_OK"

SCENE_SYNC_FINISHED = "OVER"
NAME_SYNC_FINISHED = "NAME_OVER"

CMD_CONNECT = "IOT_KEY?"


class Command(Enum):
    """
    ELRO Connects send commands copied from decompiled android app
    """

    SWITCH_TIMER = -34
    DELETE_EQUIPMENT_DETAIL = -4
    EQUIPMENT_CONTROL = 1
    INCREACE_EQUIPMENT = 2
    REPLACE_EQUIPMENT = 3
    DELETE_EQUIPMENT = 4
    MODIFY_EQUIPMENT_NAME = 5
    CHOOSE_SCENE_GROUP = 6
    CANCEL_INCREACE_EQUIPMENT = 7
    INCREACE_SCENE = 8
    MODIFY_SCENE = 9
    DELETE_SCENE = 10
    GET_DEVICE_NAME = 14
    GET_ALL_EQUIPMENT_STATUS = 15
    GET_ALL_SCENE_INFO = 18
    TIME_CHECK = 21
    INCREACE_SCENE_GROUP = 23
    MODIFY_SCENE_GROUP = 24
    SYN_DEVICE_STATUS = 29
    SYN_DEVICE_NAME = 30
    SYN_SCENE = 31
    SCENE_HANDLE = 32
    SCENE_GROUP_DELETE = 33
    MODEL_SWITCH_TIMER = 34
    MODEL_TIMER_SYN = 35
    UPLOAD_MODEL_TIMER = 36
    MODEL_TIMER_DEL = 37
    SEND_TIMEZONE = 251

    ANSWER_YES_OR_NO = 11

    # ELRO Connects receive commands
    # Reverse engineered
    DEVICE_STATUS_UPDATE = 19
    DEVICE_ALARM_TRIGGER = 25
    SCENE_STATUS_UPDATE = 26
    DEVICE_NAME_REPLY = 17

    SENCE_GROUP_DETAIL = 26
    SCENE_TYPE = 27
    SENCE_GROUP = 28


class CommandAttributes(TypedDict):
    """Base class for building command attributes for elro.api.async_process_command."""

    cmd_id: Command
    additional_attributes: dict
    attribute_transformer: Callable | None
    receive_types: list[Command]
    content_field: str
    content_sync_finished: int | str | None
    content_transformer: Callable | None


# GET_DEVICE_NAMES returns a dict[{device_id}, {device_name}]
GET_DEVICE_NAMES = CommandAttributes(
    cmd_id=Command.GET_DEVICE_NAME,
    attribute_transformer=None,
    additional_attributes={"device_ID": 0},
    receive_types=[Command.DEVICE_NAME_REPLY],
    content_field="answer_content",
    content_sync_finished="NAME_OVER",
    content_transformer=get_device_names,
)

# SET_DEVICE_NAMES returns a dict[{device_id}, {device_name}]
SET_DEVICE_NAME = CommandAttributes(
    cmd_id=Command.MODIFY_EQUIPMENT_NAME,
    attribute_transformer=set_device_name,
    additional_attributes={"device_ID": 0, "device_name": ""},
    receive_types=[Command.ANSWER_YES_OR_NO],
    content_field="answer_yes_or_no",
    content_sync_finished=2,
    content_transformer=None,
)

# SYN DEVICE_STATUS
SYN_DEVICE_STATUS = CommandAttributes(
    cmd_id=Command.SYN_DEVICE_STATUS,
    attribute_transformer=None,
    additional_attributes={"device_status": ""},
    receive_types=[Command.DEVICE_STATUS_UPDATE, Command.DEVICE_ALARM_TRIGGER],
    content_field="device_status",
    content_sync_finished="OVER",
    content_transformer=get_device_states,
)

# GET_ALL_EQUIPMENT_STATUS
GET_ALL_EQUIPMENT_STATUS = CommandAttributes(
    cmd_id=Command.GET_ALL_EQUIPMENT_STATUS,
    attribute_transformer=None,
    additional_attributes={"device_status": ""},
    receive_types=[Command.DEVICE_STATUS_UPDATE, Command.DEVICE_ALARM_TRIGGER],
    content_field="device_status",
    content_sync_finished="OVER",
    content_transformer=get_device_states,
)

# TEST_ALARM
TEST_ALARM = CommandAttributes(
    cmd_id=Command.EQUIPMENT_CONTROL,
    attribute_transformer=None,
    additional_attributes={"device_ID": 0, "device_status": "17000000"},
    receive_types=[Command.ANSWER_YES_OR_NO],
    content_field="answer_yes_or_no",
    content_sync_finished=2,
    content_transformer=None,
)

SILENCE_ALARM = CommandAttributes(
    cmd_id=Command.EQUIPMENT_CONTROL,
    attribute_transformer=None,
    additional_attributes={"device_ID": 0, "device_status": "00000000"},
    receive_types=[Command.ANSWER_YES_OR_NO],
    content_field="answer_yes_or_no",
    content_sync_finished=2,
    content_transformer=None,
)

# GET_SCENES returns a dict[{scene_id}, None]
# NOTE: If queried frequently not all data is provisioned all the time
GET_SCENES = CommandAttributes(
    cmd_id=Command.SYN_SCENE,
    attribute_transformer=None,
    additional_attributes={
        "sence_group": 0,
        "answer_content": "",
        "scene_content": "",
    },
    receive_types=[
        Command.SENCE_GROUP_DETAIL,
        Command.SCENE_TYPE,
        Command.SENCE_GROUP,
    ],
    content_field="scene_content",
    content_sync_finished="OVER",
    content_transformer=get_default,
)



"""Elro device class models."""

from enum import Enum


class DeviceType(Enum):
    """
    The DeviceType defines which kind of Elro device this is
    """

    CO_ALARM = "0000", "1000", "2000", "0008", "1008", "2008", "000E", "100E", "200E"
    GAS_ALARM = "0002", "1002", "2002", "1006", "000A", "100A", "200A"
    SMOKE_ALARM = "0001", "1001", "2001", "0009", "1009", "2009"
    WATER_ALARM = "0004", "1004", "2004", "000C", "100C", "200C", "0012", "1012", "2012"
    HEAT_ALARM = "0003", "1003", "2003", "000B", "100B", "200B", "0011", "1011", "2011"
    FIRE_ALARM = "0005", "1109", "2109", "000D", "100D", "200D", "0013", "1013", "2013"

    GUARD = "0210", "1210", "2210"  # access control
    TEMPERATURE_SENSOR = "0102", "1102", "2102"  # TH_CHECK
    DOOR_WINDOW_SENSOR = "0101", "1101", "2101"  # DOOR_CHECK

    LOCK = "1213"
    MODE_BUTTON = "0305"
    BUTTON = "0301", "1301", "2301"
    LAMP = "020A", "120A", "220A"
    SOCKET = "0200", "1200", "2200"
    TWO_SOCKET = "0201", "1201", "2201"
    VALVE = "0208", "1208", "2208"
    CURTAIN = "0209", "1209", "2209"
    TEMP_CONTROL = "0215", "1215", "2215"
    DIMMING_MODULE = "0214", "1214", "2214"

    SOS_KEY = "0300", "1300", "2300"
    PIR_CHECK = "0100", "1100", "2100"

    def __new__(cls, *values):
        obj = object.__new__(cls)
        # first value is canonical value
        obj._value_ = values[0]
        for other_value in values[1:]:
            cls._value2member_map_[other_value] = obj  # pylint: disable=maybe-no-member
        obj._all_values = values
        return obj

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}.{self._name_}: "  # pylint: disable=maybe-no-member
            f"{', '.join([repr(v) for v in self._all_values])}>"
        )


# Represent alarm device types
ALARM_CO = "CO_ALARM"
ALARM_FIRE = "FIRE_ALARM"
ALARM_HEAT = "HEAT_ALARM"
ALARM_SMOKE = "SMOKE_ALARM"
ALARM_WATER = "WATER_ALARM"

# Represent device state attributes
ATTR_BATTERY_LEVEL = "battery"
ATTR_DEVICE_TYPE = "device_type"
ATTR_DEVICE_STATE = "device_state"
ATTR_SIGNAL = "signal"

# Represent the state of a device.
DEVICE_STATE = {
    "12": "FAULT",
    "15": "SILENCE",
    "17": "TEST ALARM",
    "19": "FIRE ALARM",
    "1B": "SILENCE",
    "BB": "TEST ALARM",
    "55": "ALARM",  # OPEN window sensor
    "AA": "NORMAL",  # CLOSED window sensor
    "FE": "UNKNOWN",
    "FF": "OFFLINE",
}

STATE_ALARM = "ALARM"
STATE_FAULT = "FAULT"
STATE_FIRE_ALARM = "FIRE ALARM"
STATE_NORMAL = "NORMAL"
STATE_OFFLINE = "OFFLINE"
STATE_SILENCE = "SILENCE"
STATE_TEST_ALARM = "TEST ALARM"
STATE_UNKNOWN = "UNKNOWN"

STATES_ON = (STATE_ALARM, STATE_FIRE_ALARM, STATE_TEST_ALARM)
STATES_OFFLINE = (STATE_OFFLINE,)



"""Example code to demonstrate the elro connects PI API with a firealarms."""

import asyncio
import logging
import sys

from elro.api import K1
from elro.command import (
    GET_SCENES,
    SET_DEVICE_NAME,
    SYN_DEVICE_STATUS,
    GET_DEVICE_NAMES,
    GET_ALL_EQUIPMENT_STATUS,
    TEST_ALARM,
    SILENCE_ALARM,
)
from elro.utils import update_state_data

INTERVAL = 5


class AlarmDemo(K1):
    """Class to demonstrate Elro connects fire alarms."""

    async def async_demo(self) -> None:
        """Main routine to demonstrate the API code."""
        logging.basicConfig(level=logging.DEBUG)
        await self.async_connect()

        print("Demo GET_SCENES")
        print(await self.async_process_command(GET_SCENES, sence_group=0))

        print("Demo SYN_DEVICE_STATUS with GET_DEVICE_NAMES")
        data = await self.async_process_command(SYN_DEVICE_STATUS)
        names = await self.async_process_command(GET_DEVICE_NAMES)
        update_state_data(data, names)
        print(data)

        print("Demo GET_ALL_EQUIPMENT_STATUS with GET_DEVICE_NAMES")
        data = await self.async_process_command(GET_ALL_EQUIPMENT_STATUS)
        names = await self.async_process_command(GET_DEVICE_NAMES)
        update_state_data(data, names)
        print(data)
        await asyncio.sleep(INTERVAL)

        # Update name of device 1
        current_name = data[1]["name"] if data else ""
        print(f"Set name demo device 1. Current name is '{current_name}'")
        new_name = "Changed name"
        await self.async_process_command(
            SET_DEVICE_NAME, device_ID=1, device_name=new_name
        )
        names = await self.async_process_command(GET_DEVICE_NAMES)
        update_state_data(data, names)
        updated_name = data[1]["name"] if data else ""
        print(f"Set name demo device 1. New name is now '{updated_name}'!")
        await asyncio.sleep(INTERVAL)
        print("Restore old name")
        await self.async_process_command(
            SET_DEVICE_NAME, device_ID=1, device_name=current_name
        )
        names = await self.async_process_command(GET_DEVICE_NAMES)
        update_state_data(data, names)
        updated_name = data[1]["name"] if data else ""
        print(f"Set name demo device 1. Name is again '{updated_name}'!")
        await self.async_disconnect()
        await asyncio.sleep(INTERVAL)
        update_nr = 0
        while update_nr < 2:
            try:
                update_nr += 1
                print(f"Demo status update {update_nr}...")
                # Start a new session
                await self.async_connect()
                status_update = await self.async_process_command(
                    GET_ALL_EQUIPMENT_STATUS
                )
                update_state_data(data, status_update)
                print(data)
                await asyncio.sleep(INTERVAL)
            except asyncio.exceptions.CancelledError as canceled_error:
                print(f"A cancelled error occcured! {canceled_error}")
                break
            except asyncio.exceptions.InvalidStateError as invalid_state_error:
                print(f"An invalid state error occcured! {invalid_state_error}")
            except asyncio.exceptions.TimeoutError as timeout_error:
                print(f"A timeout error occcured! {timeout_error}")
            finally:
                # Close the connection if needed
                await self.async_disconnect()

        # Test alarm (assuming there are 3 fire alarms)
        # Be aware they cannot be fired alle together, silence an alarm befor testing the next alarm.
        await self.async_connect()
        print("Test alarm 1")
        await self.async_process_command(TEST_ALARM, device_ID=1)
        await asyncio.sleep(4)
        print("Test alarm 2")
        await self.async_process_command(SILENCE_ALARM, device_ID=1)
        await self.async_process_command(TEST_ALARM, device_ID=2)
        await asyncio.sleep(4)
        print("Test alarm 3")
        await self.async_process_command(SILENCE_ALARM, device_ID=2)
        await self.async_process_command(TEST_ALARM, device_ID=3)
        await asyncio.sleep(4)
        await self.async_process_command(SILENCE_ALARM, device_ID=3)
        await self.async_disconnect()
        print("Demo completed")


if __name__ == "__main__":
    # argv: 1 = IP_ADDRESS, 2 = API_KEY
    k1_hub = AlarmDemo(sys.argv[1], sys.argv[2])
    asyncio.run(k1_hub.async_demo())



"""Setup lib-elro-connects."""

import os
import re
import sys

from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    path = os.path.join(os.path.dirname(__file__), package, "__init__.py")
    with open(path, "rb") as file:
        init_py = file.read().decode("utf-8")
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


if sys.version_info < (3, 8):
    sys.exit("Sorry, Python versions < 3.8 are not supported")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = []

setup(
    name="lib-elro-connects",
    version=get_version("elro"),
    license="MIT Licence",
    author="Jan Bouwhuis",
    author_email="jan@jbsoft.nl",
    description="Elro Connects P1 API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jbouwh/lib-elro-connects",
    # package_dir={"": ""},
    packages=find_packages(where="elro"),
    data_files=[
        (
            "share/lib-elro-connects",
            [
                "examples/fire_alarm_demo.py",
            ],
        )
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Home Automation",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    install_requires=install_requires,
    scripts=[],
)



"""Elro connects API."""

from __future__ import annotations

import asyncio
import json
import logging
from typing import Mapping, cast, Any, TypedDict

from elro.command import (
    Command,
    CommandAttributes,
    ACK_APP,
    CMD_CONNECT,
)
from elro.utils import (
    validate_json,
)

ATTR_BIND = "BIND"
ATTR_KEY = "KEY"
ATTR_NAME = "NAME"

APP_ID = 0

TIMEOUT = 10
INTERVAL = 5
UDP_PORT_NO = 1025

_LOGGER = logging.getLogger(__name__)


class K1UDPHandler(asyncio.BaseProtocol):
    """UDP test class."""

    def __init__(self, message, on_con_lost, datagram_data):
        self.message = message
        self.on_con_lost = on_con_lost
        self._transport = None
        self.datagram_data = datagram_data
        self.last_exc = None

    def connection_made(self, transport):
        """Connection made."""
        self._transport = transport
        self._transport.sendto(self.message)

    def datagram_received(self, data, addr):
        """Datagram reveived."""
        if not self.datagram_data.done():
            self.datagram_data.set_result((data, addr))

    def close_connection(self):
        """Close the connection."""
        self._transport.close()

    def error_received(self, exc):
        """Error received."""
        print("Error received:", exc)
        self.last_exc = exc

    def connection_lost(self, exc):
        """Connection lost."""
        self.last_exc = exc
        self.on_con_lost.set_result(True)
        if not self.datagram_data.done():
            self.datagram_data.set_result((None, None))


class K1:
    """API class to Elro connects K1 adapter."""

    _lock = asyncio.Lock()
    _loop: asyncio.AbstractEventLoop | None = None

    class K1ConnectionError(Exception):
        """K1 exception class."""

        def __init__(self, message: str = "K1 connection error"):
            self.message = message
            super().__init__(self.message)

    class Reponse(TypedDict):
        """API Response class"""

        received: int

    def __init__(self, ipaddress: str, k1_id: str, port: int = 1025) -> None:
        """Initialize the module."""
        self._transport = None
        self._protocol = None
        self._remoteaddress = (ipaddress, port)
        self._k1_id = k1_id
        self._session: dict[str, str] = {}
        self._msg_id = 0

    async def async_connect(self) -> None:
        """Connect to the K1 hub."""

        def _store_session(self, data: str) -> None:
            """Stores the session details."""
            if data:
                self._session = {}
                for line in data.rstrip().split("\n"):
                    key, value = line.strip().split(":")
                    self._session[key] = value

        self._loop = asyncio.get_running_loop()
        if not self._loop:
            return
        on_conn_lost, datagram_data = (
            self._loop.create_future(),
            self._loop.create_future(),
        )
        payload = (CMD_CONNECT + self._k1_id).encode("utf-8")
        await self._lock.acquire()
        try:
            self._transport, self._protocol = await self._loop.create_datagram_endpoint(  # type: ignore
                lambda: K1UDPHandler(payload, on_conn_lost, datagram_data),
                remote_addr=self._remoteaddress,
            )
            await asyncio.wait_for(datagram_data, 15.0)
            if data := datagram_data.result():
                _store_session(self, data[0].decode("utf-8"))
                return
            raise K1.K1ConnectionError(
                "No data received, cannot connect to "
                f"hub {self._remoteaddress[0]} with id {self._k1_id}."
            )
        except (ValueError, asyncio.TimeoutError) as exception:
            raise K1.K1ConnectionError(
                "Not received the expected result, cannot connect to "
                f"hub {self._remoteaddress[0]} with id {self._k1_id}."
                f" {exception.args}"
            ) from exception
        finally:
            self._lock.release()

    async def async_disconnect(self) -> None:
        """Disconnect from the K1 hub."""
        if not self._protocol:
            return
        await self._lock.acquire()
        try:
            if self._transport and not self._protocol.on_con_lost.done():
                self._transport.close()
                self._transport = None
                self._protocol = None
                self._session = {}
        finally:
            self._lock.release()

    def _prepare_command(self, command_data: dict) -> bytes:
        """
        Construct a valid message from data
        :param data: A string containing data to be send to the K1
        :return: A json message
        """
        self._msg_id += 1

        command = {
            "msgId": self._msg_id,
            "action": "appSend",
            "params": {
                "devTid": self._session[ATTR_NAME],
                "ctrlKey": self._session[ATTR_KEY],
                "appTid": 1,
                "data": command_data,
            },
        }
        return json.dumps(command).encode("utf-8")

    async def async_process_command(
        self,
        attributes: CommandAttributes,
        **argv: int | str,
    ) -> dict[int, dict[str, Any]] | None:
        """Get device names."""
        iteration = 0
        if (
            not self._protocol
            or not self._transport
            or not self._loop
            or not self._session
        ):
            raise K1.K1ConnectionError("Not connected to a K1 hub.")

        if attributes["attribute_transformer"]:
            attributes["attribute_transformer"](argv)
        command_data = {
            "cmdId": attributes["cmd_id"].value,
        }
        command_data.update(attributes["additional_attributes"])
        if argv:
            command_data.update(cast(Mapping[str, Any], argv))
        command = self._prepare_command(command_data)
        contentlist = []
        await self._lock.acquire()
        try:
            self._protocol.datagram_data = self._loop.create_future()
            self._transport.sendto(command)
            while True:
                # Run loop until last item
                await asyncio.wait_for(self._protocol.datagram_data, TIMEOUT)
                if raw_data := self._protocol.datagram_data.result():
                    iteration += 1
                    _LOGGER.debug(
                        "command attributes: %s received[%s]: %s",
                        command_data,
                        iteration,
                        raw_data[0].decode("utf-8").strip()
                        if raw_data[0] is not None
                        else None,
                    )
                    self._protocol.datagram_data = self._loop.create_future()
                    data = validate_json(raw_data[0])
                    params = data["params"]
                    cmd_id = params["data"]["cmdId"]
                    content = params["data"].get(attributes["content_field"], "")
                    if Command(cmd_id) in attributes["receive_types"]:
                        self._transport.sendto(ACK_APP.encode("utf-8"))
                        if content == attributes["content_sync_finished"]:
                            break
                        contentlist.append(params["data"])

                else:
                    self._transport.sendto(ACK_APP.encode("utf-8"))
                    break
        except (ValueError, asyncio.TimeoutError, asyncio.CancelledError) as exception:
            raise K1.K1ConnectionError(
                "Not received the expected result, cannot connect to "
                f"hub {self._remoteaddress[0]} with id {self._k1_id}."
                f" {exception.args}"
            ) from exception
        finally:
            self._lock.release()
        return (
            attributes["content_transformer"](contentlist)
            if attributes["content_transformer"] is not None
            else None
        )



"""Utilities to support Elro Connects P1 API."""

from __future__ import annotations

import json
import logging
import collections
from typing import Any

from elro.device import DeviceType, DEVICE_STATE


# From the ByteUtil class, needed by CRC_maker
AUCHCRCHI = (
    b"\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x00\xc1"
    b"\x81@\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@\x00\xc1\x81@\x01"
    b"\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A"
    b"\x00\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x01\xc0"
    b"\x80A\x00\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@\x01"
    b"\xc0\x80A\x00\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@"
    b"\x00\xc1\x81@\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1"
    b"\x81@\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x00"
    b"\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x01\xc0\x80A"
    b"\x00\xc1\x81@\x01\xc0\x80A\x00\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x01\xc0"
    b"\x80A\x00\xc1\x81@\x00\xc1\x81@\x01\xc0\x80A\x00\xc1\x81@\x01\xc0\x80A\x01"
    b"\xc0\x80A\x00\xc1\x81@"
)
# From the ByteUtil class, needed by CRC_maker
AUCHCRCLO = (
    b"\x00\xc0\xc1\x01\xc3\x03\x02\xc2\xc6\x06\x07\xc7\x05\xc5\xc4\x04\xcc\x0c\r"
    b"\xcd\x0f\xcf\xce\x0e\n\xca\xcb\x0b\xc9\t\x08\xc8\xd8\x18\x19\xd9\x1b\xdb\xda"
    b"\x1a\x1e\xde\xdf\x1f\xdd\x1d\x1c\xdc\x14\xd4\xd5\x15\xd7\x17\x16\xd6\xd2\x12"
    b"\x13\xd3\x11\xd1\xd0\x10\xf001\xf13\xf3\xf226\xf6\xf77\xf554\xf4<\xfc\xfd="
    b"\xff?>\xfe\xfa:;\xfb9\xf9\xf88(\xe8\xe9)\xeb+*\xea\xee./\xef-\xed\xec,\xe4$%"
    b"\xe5'\xe7\xe6&\"\xe2\xe3#\xe1! \xe0\xa0`a\xa1c\xa3\xa2bf\xa6\xa7g\xa5ed\xa4l"
    b"\xac\xadm\xafon\xae\xaajk\xabi\xa9\xa8hx\xb8\xb9y\xbb{z\xba\xbe~\x7f\xbf}\xbd"
    b"\xbc|\xb4tu\xb5w\xb7\xb6vr\xb2\xb3s\xb1qp\xb0P\x90\x91Q\x93SR\x92\x96VW\x97U"
    b"\x95\x94T\x9c\\]\x9d_\x9f\x9e^Z\x9a\x9b[\x99YX\x98\x88HI\x89K\x8b\x8aJN\x8e"
    b"\x8fO\x8dML\x8cD\x84\x85E\x87GF\x86\x82BC\x83A\x81\x80@"
)


def get_string_from_ascii(input_string):
    """
    This function is reversed engineered and translated to python
    based on the CoderUtils class in the ELRO Android app

    :param input: A hex string
    :return: A string
    """
    try:
        if len(input_string) != 32:
            raise ValueError("input {} has not the required length of 32.")

        byt = bytearray.fromhex(input_string)
        name = "".join(map(chr, byt))
        name = name.replace("@", "").replace("$", "")
    except ValueError as error:
        raise error

    return name


def get_ascii(input_string: str) -> str | None:
    """Encode a string to hex for API use."""
    #
    # This function is partially reversed engineered and translated to python
    # based on the CoderUtils class in the ELRO Android app

    # :param input: A string
    # :return: A hex string

    # Pattern("^[_\-a-zA-Z0-9 ]*$"

    countf = 0

    countf = 15 - len(input_string.encode("GBK"))

    if countf < 0:
        logging.error("Input is to long '%s'", input_string)
        return None

    str_whitespace = ""
    for _ in range(countf):
        str_whitespace += "@"

    new_name = str_whitespace + input_string + "$"

    # This is the original code and thus where the python code differs.
    # Because this is not used, GBK encoding is probably not fully supported

    # byte[] nameBt = new byte[16];
    # try {
    #    nameBt = newname.getBytes("GBK");
    # } catch (UnsupportedEncodingException e) {
    #    e.printStackTrace();
    # }

    # String ds = "";
    # for(int i=0;i<nameBt.length;i++){
    #    String str = ByteUtil.convertByte2HexString(nameBt[i]);
    #    ds+=str;
    # }

    # These two functions are also used, from the ByteUtils
    # public static String convertByte2HexString(byte b) {
    #    char u8 = convertByte2Uint8(b);
    #    String buf = Integer.toHexString(u8);
    #    if (buf.length() == 1) {
    #        buf = '0' + buf;
    #    }
    #    return buf;
    # }

    # public static char combine2bytesToU16(byte high, byte low) {
    #    char highU8 = convertByte2Uint8(high);
    #    char lowU8 = convertByte2Uint8(low);
    #    return (char) (highU8 << 8 | lowU8);
    # }

    return new_name.encode("GBK").hex()


def crc_maker(msg):
    """
    This function is reversed engineered and translated to python
    based on the ByteUtil class in the ELRO Android app

    :param input: The string to create a CRC for
    :return: A CRC string
    """
    crc_hi_1 = 0
    crc_lo_1 = 0
    msg_length = len(msg)
    index = 0
    crc_index = 0
    uch_crc_hi = 0xFF
    uch_crc_lo = 0xFF

    while index < msg_length:
        char_code = ord(msg[index])

        crc_index = uch_crc_hi ^ char_code
        uch_crc_hi = uch_crc_lo ^ AUCHCRCHI[crc_index]
        uch_crc_lo = AUCHCRCLO[crc_index]

        index = index + 1

    crc_lo_1 = uch_crc_hi
    crc_lo_2 = int(crc_lo_1)
    crc_lo_3 = hex(crc_lo_2)[2:]
    if len(crc_lo_3) < 2:
        crc_lo_3 = "0" + crc_lo_3

    crc_hi_1 = uch_crc_lo
    crc_hi_2 = int(crc_hi_1)
    crc_hi_3 = hex(crc_hi_2)[2:]
    if len(crc_hi_3) < 2:
        crc_hi_3 = "0" + crc_hi_3

    return f"{crc_hi_3.upper()}{crc_lo_3.upper()}"


def crc_maker_char(msg):
    """
    This function is reversed engineered and translated to python
    based on the ByteUtil class in the ELRO Android app

    :param input: The string to create a CRC for
    :return: A CRC string
    """

    crc_hi_1 = 0
    crc_lo_1 = 0
    index = 0
    crc_index = 0
    uch_crc_hi = 0xFF
    uch_crc_lo = 0xFF

    msg_length = int(len(msg) / 2)
    content = []

    for _ in range(msg_length):
        val = chr(int((msg[0:2]), 16))
        content.append(val)
        msg = msg[2:]

    while index < msg_length:
        char_code = ord(content[index])

        crc_index = uch_crc_hi ^ char_code
        uch_crc_hi = uch_crc_lo ^ AUCHCRCHI[crc_index]
        uch_crc_lo = AUCHCRCLO[crc_index]

        index = index + 1

    crc_lo_1 = uch_crc_hi
    crc_lo_2 = int(crc_lo_1)
    crc_lo_3 = hex(crc_lo_2)[2:]
    if len(crc_lo_3) < 2:
        crc_lo_3 = "0" + crc_lo_3

    crc_hi_1 = uch_crc_lo
    crc_hi_2 = int(crc_hi_1)
    crc_hi_3 = hex(crc_hi_2)[2:]
    if len(crc_hi_3) < 2:
        crc_hi_3 = "0" + crc_hi_3

    return f"{crc_hi_3.upper()}{crc_lo_3.upper()}"


def get_eq_crc(devices):
    """
    Builds a CRC string based on device id and device status. This function is reverse engineered
    and translated to python. It is based on the CoderUtils class in the elro app :param devices:.
    A dictionary of devices statuses, where the id of the device is the index of the dict
    """
    sorted_devices = collections.OrderedDict(sorted(devices.items()))
    list_length = int(list(sorted_devices.keys())[-1])

    status_crc = ""
    for i in range(list_length + 1):
        if i + 1 in sorted_devices:
            status_crc += crc_maker_char(sorted_devices[i + 1])
        elif i < (list_length):
            status_crc += "0000"

    num = ""
    list_length_for_hex = hex((list_length * 2 + 2))[2:]
    if len(list_length_for_hex) < 4:
        i = 4 - len(list_length_for_hex)
        num = list_length_for_hex.rjust(i + len(list_length_for_hex), "0")
    else:
        num = list_length_for_hex

    return num + status_crc


def update_state_data(
    data: dict[int, dict[str, Any]] | None,
    data_update: dict[int, dict[str, Any]] | None,
) -> None:
    "Update the state."
    if data_update is None or data is None:
        return
    for key in data_update.keys():
        if not key in data:
            data[key] = data_update[key]
        else:
            data[key].update(data_update[key])


def get_device_names(content: list) -> dict:
    """Return device names."""
    # answer_content
    return {
        int(data["answer_content"][0:4], 16): {
            "name": get_string_from_ascii(data["answer_content"][4:])
        }
        for data in content
    }


def set_device_name(argv: dict) -> None:
    """Convert the device_name attribute to a hex representation including crc."""
    if device_name := argv.get("device_name"):
        device_name_hex = get_ascii(device_name)
        crc = crc_maker(device_name)
        argv["device_name"] = f"{device_name_hex}{crc}"
    else:
        raise ValueError("Value for device_name is not set!")


def get_device_states(content: list) -> dict:
    """Return device states."""

    return {
        hexdata["device_ID"]: {
            "device_type": DeviceType(hexdata["device_name"]).name,
            "signal": int(hexdata["device_status"][0:2], 16),
            "battery": int(hexdata["device_status"][2:4], 16),
            "device_state": DEVICE_STATE.get(hexdata["device_status"][4:6], "n/a"),
            "device_status_data": hexdata,
        }
        for hexdata in content
    }


def get_default(content: list) -> dict:
    """Return content from ascii."""
    index = 0
    return_dict = {}
    for line in content:
        return_dict[index] = line
        index += 1

    return return_dict


def validate_json(raw_data: bytes) -> dict:
    """Process the JSON basis response, work-a-round synatx errors."""
    json_string = raw_data.decode("utf-8")
    try:
        data = json.loads(json_string)
    except json.decoder.JSONDecodeError:
        # Try correct the malformed JSON status string missing to closing brackets
        data = json.loads(json_string + "}}")
    return data
